/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

#ifndef APACHE_BUFF_H
#define APACHE_BUFF_H

#ifdef __cplusplus
extern "C" {
#endif

#ifdef APR_HAVE_STDARG_H
#include <stdarg.h>
#endif
#include "apr_lib.h"
#include "apr_general.h"
#ifdef APACHE_XLATE
#include "apr_xlate.h"
#endif
    
/*
    A BUFF is an i/o object which can be used in any of the following
    output modes:

    blocking, buffered
    blocking, buffered, HTTP-chunked
    blocking, unbuffered
    blocking, unbuffered, HTTP-chunked
    non-blocking, unbuffered
    non-blocking, unbuffered, HTTP-chunked

    In all the blocking modes, a bwrite(fb, buf, len) will write less
    than len only in an error state.  The error may be deferred until
    the next use of fb.

    In the non-blocking, chunked mode, the caller of bwrite() makes a
    guarantee that if a partial write occurs, they will call back later
    with at least as many bytes to write -- prior to disabling chunking.
    This is a protocol correctness requirement -- the chunk length may
    already have hit the wire, and is in essence "committed".

    bputc, bputs, bvputs, and bprintf are supported only in the buffered
    modes.

    The following input modes are supported:

    blocking, buffered
    blocking, unbuffered
    non-blocking, unbuffered
*/

/* Reading is buffered */
#define B_RD     (1)
/* Writing is buffered */
#define B_WR     (2)
#define B_RDWR   (3)
/* At end of file, or closed stream; no further input allowed */
#define B_EOF    (4)
/* No further output possible */
#define B_EOUT   (8)
/* A read error has occurred */
#define B_RDERR (16)
/* A write error has occurred */
#define B_WRERR (32)
#ifdef B_ERROR  /* in SVR4: sometimes defined in /usr/include/sys/buf.h */
#undef B_ERROR
#endif
#define B_ERROR (48)
/* Use chunked writing */
#define B_CHUNK (64)
/* bflush() if a read would block */
#define B_SAFEREAD (128)
/* caller expects non-blocking behaviour */
#define B_NONBLOCK (512)

/* TODO: implement a ebcdic/ascii conversion layers */

typedef struct buff_struct BUFF;

#ifdef APACHE_XLATE
typedef struct b_read_ops {
    apr_status_t (*bread)(BUFF *, void *, apr_size_t, apr_ssize_t *);
} b_read_ops;

typedef struct b_write_ops {
    apr_status_t (*bwrite)(BUFF *, const void *, apr_size_t, apr_ssize_t *);
} b_write_ops;
#endif

struct buff_struct {
    int flags;			/* flags */
    apr_status_t saved_errno;	/* saved errno */
    apr_status_t berrno;         /* errno for the functions that don't return
                                   an apr_status_t. This is separate from
                                   saved_errno. A broken call, e.g. ap_bgets
                                   on an unbuffered stream, shouldn't change
                                   the saved error from I/O */
    unsigned char *inptr;	/* pointer to next location to read */
    apr_ssize_t incnt;		/* number of bytes left to read from input buffer;
				 * always 0 if had a read error  */
    int outchunk;               /* location of chunk header when chunking */
    int outcnt;			/* number of byte put in output buffer */
    unsigned char *inbase;
    unsigned char *outbase;
    int bufsiz;
    int chunk_overcommit;	/* when we start a chunk and get a partial write we
				 * keep track of the #remaining bytes in the chunk
				 * here
				 */
    void (*error) (BUFF *fb, int op, void *data, apr_status_t errnum);
    void *error_data;
    long int bytes_sent;	/* number of bytes actually written */

    apr_pool_t *pool;

    apr_socket_t *bsock;
#ifdef APACHE_XLATE
    b_read_ops b_read_ops;      /* some read operations are replaced when 
                                 * translation is on */
    b_write_ops b_write_ops;    /* some write operations are replaced when
                                 * translation is on */
    apr_xlate_t *write_xlate;    /* translation handle to use on write operations
                                 * (or NULL if no translation is to be performed) */
    int write_xlate_sb;         /* flag: is write translation single-byte-only? */
    apr_xlate_t *read_xlate;     /* translation handle to use on read operations
                                 * (or NULL if no translation is to be performed) */
    char *xbuf;                 /* buffer to use temporarily while translating */
    size_t xbufsize;            /* size of xbuf buffer */
    unsigned char last_getc_raw; /* last character retrieved via ap_bgetc(), before
                                  * translation; because read translation may be
                                  * different than write translation, we can't use
                                  * a write translation handle to put back what
                                  * we previously translated with the read
                                  * translation handle; thus, we need the
                                  * untranslated version in case ap_bungetc() is
                                  * called */
    char partial_write[8];       /* incomplete output character; must hold the largest
                                    character in any charset*/
    size_t partial_write_len;    /* bytes held in partial_write */
#endif /*APACHE_XLATE*/
};

/* Options to bset/getopt */
#define BO_BYTECT (1)
/* XXX: timeouts should probably be redefined to be in microseconds to match APR */
/* timeout values have the same semantics as those documented for timeouts
   in apr_socket.h... that is:
   timeout < 0 is blocking infinite/no-timeout
   timeout == 0 is non-blocking
   timeout > 0 is blocking with timeout after the indicated number of seconds
*/
#define BO_TIMEOUT (2)
#define BO_ERROR (3)
#ifdef APACHE_XLATE
/* BO_WXLATE - translate-charset-on-write setting; if value is NULL, no
 *             translation is performed; otherwise, value is apr_xlate_t * and
 *             translation will be performed
 */
#define BO_WXLATE (4)
/* BO_RXLATE - translate-charset-on-read setting; if value is NULL, no
 *             translation is performed; otherwise, value is apr_xlate_t * and
 *             translation will be performed
 */
#define BO_RXLATE (5)
#endif /*APACHE_XLATE*/

/* Stream creation and modification */
API_EXPORT(BUFF *) ap_bcreate(apr_pool_t *p, int flags);

API_EXPORT(void) ap_bpush_socket(BUFF *fb, apr_socket_t *sock);
API_EXPORT(void) ap_bpop_socket(apr_socket_t **sock, BUFF *fb);

API_EXPORT(apr_status_t) ap_bsetopt(BUFF *fb, int optname, const void *optval);
API_EXPORT(apr_status_t) ap_bgetopt(BUFF *fb, int optname, void *optval);
API_EXPORT(int) ap_bsetflag(BUFF *fb, int flag, int value);
API_EXPORT(apr_status_t) ap_bclose(BUFF *fb);

#define ap_bgetflag(fb, flag)	((fb)->flags & (flag))

/* Error handling */
API_EXPORT(void) ap_bonerror(BUFF *fb,
                             void (*error) (BUFF *, int, void *, apr_status_t),
                             void *data);
API_EXPORT(apr_status_t) ap_berror(BUFF *fb);

/* I/O */
/* read bytes into the specified buffer.
 * *bytes_read is set to the number of bytes read.  it is possible for
 * bytes to have been read even if an error (or APR_EOF) is returned.
 */
API_EXPORT(apr_status_t) ap_bread_core(BUFF *fb, void *buf, apr_size_t nbyte,
                                 apr_ssize_t *bytes_read);
API_EXPORT(int) ap_bgets(char *s, int n, BUFF *fb);
API_EXPORT(int) ap_blookc(BUFF *fb);
/* write bytes from the specified buffer.
 * *bytes_written is set to the number of bytes written.  it is possible
 * for both an error to be returned, and bytes to have been written.
 */
API_EXPORT(apr_status_t) ap_bwrite_core(BUFF *fb, const void *buf, apr_size_t nbyte,
                                  apr_ssize_t *bytes_written);
#ifdef APACHE_XLATE
#define ap_bwrite(fb,buf,nbyte,bytes_written) \
(fb)->b_write_ops.bwrite(fb,buf,nbyte,bytes_written)
#define ap_bread(fb,buf,nbyte,bytes_read) \
(fb)->b_read_ops.bread(fb,buf,nbyte,bytes_read)
#else /*APACHE_XLATE*/
#define ap_bwrite(fb,buf,nbyte,bytes_written) \
ap_bwrite_core(fb,buf,nbyte,bytes_written)
#define ap_bread(fb,buf,nbyte,bytes_read) \
ap_bread_core(fb,buf,nbyte,bytes_read)
#endif /*APACHE_XLATE*/
    
API_EXPORT(apr_status_t) ap_bflush(BUFF *fb);
API_EXPORT(apr_status_t) ap_bshutdown(BUFF *fb, int how);
API_EXPORT(int) ap_bputs(const char *x, BUFF *fb);
API_EXPORT_NONSTD(int) ap_bprintf(BUFF *fb, const char *fmt,...)
				__attribute__((format(printf,2,3)));
API_EXPORT(int) ap_vbprintf(BUFF *fb, const char *fmt, va_list vlist);

/* ### ap_bvputs is a misnomer, since it doesn't take a va_list */
#define ap_bvputs       ap_bputstrs
API_EXPORT_NONSTD(int) ap_bputstrs(BUFF *fb, ...);
API_EXPORT(int) ap_vbputstrs(BUFF *fb, va_list va);

/* Internal routines */
API_EXPORT(int) ap_bflsbuf(int c, BUFF *fb);
API_EXPORT(int) ap_bfilbuf(BUFF *fb);
/* bflush() now if a read would have to block */
API_EXPORT(void) ap_bhalfduplex(BUFF *fb);


#ifndef APACHE_XLATE
  
#define ap_bgetc(fb)   ( ((fb)->incnt == 0) ? ap_bfilbuf(fb) : \
		    ((fb)->incnt--, *((fb)->inptr++)) )

/* can only unget a single character that was read by ap_bgetc */
#define ap_bungetc(c, fb)  ((fb)->incnt++, *(--(fb)->inptr) = (c))
			

#define ap_bputc(c, fb) ((((fb)->flags & (B_EOUT|B_WRERR|B_WR)) != B_WR || \
		     (fb)->outcnt == (fb)->bufsiz) ? ap_bflsbuf(c, (fb)) : \
		     ((fb)->outbase[(fb)->outcnt++] = (c), 0))

#else /*APACHE_XLATE*/
 
/* can only unget a single character that was read by ap_bgetc */
#define ap_bungetc(c, fb)  ((fb)->incnt++, --(fb)->inptr)
			
#define ap_bgetc(fb)   ( ((fb)->incnt == 0) ? (fb)->last_getc_raw = ap_bfilbuf(fb) : \
                    ((fb)->incnt--, (fb)->last_getc_raw = *((fb)->inptr++), \
                    ((fb)->read_xlate)\
                    ?apr_xlate_conv_byte((fb)->read_xlate,(fb)->last_getc_raw): \
                    (fb)->last_getc_raw) )
 
#define ap_bputc(c, fb) ((((fb)->flags & (B_EOUT|B_WRERR|B_WR)) != B_WR || \
                     (fb)->outcnt == (fb)->bufsiz) ? ap_bflsbuf(c, (fb)) : \
                     ((fb)->outbase[(fb)->outcnt++] = ((fb)->write_xlate)\
                     ?apr_xlate_conv_byte((fb)->write_xlate,(unsigned char)c):(c), 0))
 
#endif /*APACHE_XLATE*/

#ifdef __cplusplus
}
#endif

#endif	/* !APACHE_BUFF_H */
