/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

#include "mod_cache.h"
#include "apr_strings.h"
#include "http_conf_globals.h"
#include "http_log.h"
#include "http_main.h"

#if defined(ULTRIX_BRAIN_DEATH) || defined(SINIX_D_RESOLVER_BUG)
extern char *mktemp(char *template);
#endif 

static cache_req *create_cache_el(ap_cache_handle_t *h, const char *name)
{
    cache_req *cq = apr_pcalloc(h->pool, sizeof(cache_req));
    memset(cq, '\0', sizeof(cache_req));
    cq->el.cache = h;
    cq->el.name = name;
    cq->state = FC_READ_NONE;
    return cq;
}

apr_status_t file_cache_element(ap_cache_handle_t *h, const char *name, ap_cache_el **el_in,
                               ap_cache_query flag)
{
    cache_server_conf *conf = (cache_server_conf *) ap_get_module_config(h->server->module_config, &cache_module);

    switch(flag)
    {
    case AP_CACHE_CREATE:
    {
        cache_req *cq = create_cache_el(h, name);

        *el_in = NULL;
        cq->state |= FC_TEMP_FILE;
        /* open temporary file */
#define TMPFILESTR    "/tmpXXXXXX"
        if (conf->root == NULL)
            return errno;
        cq->tempfile = apr_pstrcat(h->pool, conf->root, TMPFILESTR, NULL);
#undef TMPFILESTR
        if(!mktemp(cq->tempfile))
            return errno;
        (*el_in) = (ap_cache_el*)cq;
        return APR_SUCCESS;
    }
    case AP_CACHE_SEEK:
    {
        apr_file_t *fp = NULL;
        apr_status_t ret;
        char *data = data_file(h, name);
        *el_in = NULL;
        ret = apr_open(&fp, data, APR_WRITE | APR_READ | APR_BINARY, 0, h->pool);
        if (ret == APR_SUCCESS)
        {
            cache_req *cq = create_cache_el(h, name);
            cq->fp = fp;
            cq->datafile = data;
            cq->hdrsfile = header_file(h, name);
            if(file_cache_read_hdrs(cq))
            {
                apr_time_t now = apr_now();
                if(cq->expire > now)
                {
                    (*el_in) = (ap_cache_el *)cq;
                    ret = APR_SUCCESS;
                }
                else
                {
                    ap_log_error(APLOG_MARK, APLOG_ERR, 0, h->server,
                                 "cache: cache entry has expired, seek failed");
                }
            }
        }
        /* XXX: is this the only case we care to report? */
        else if(APR_STATUS_IS_EACCES(ret))
            ap_log_error(APLOG_MARK, APLOG_ERR, ret, h->server,
                         "cache: cache entry couldn't be opened, perhaps permission error");
        return ret;
    }
    case AP_CACHE_REMOVE:
    {
        apr_status_t ret = APR_SUCCESS;
        char *data = data_file(h, name);
        char *hdrs = header_file(h, name);
        
        if(unlink(data) == -1)
        {
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, h->server,
                         "cache: error deleting data file %s", data);
            ret = errno;
        }
        if(unlink(hdrs) == -1)
        {
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, h->server,
                         "cache: error deleting header file %s", hdrs);
            ret = errno;
        }
        return ret;
    }
    default:
        break;
    }
    
    return APR_ENOTIMPL;
}

apr_status_t file_cache_el_final(ap_cache_el *el)
{
    cache_req *cq = (cache_req *)el;
    cache_server_conf *conf = (cache_server_conf *)
        ap_get_module_config(cq->el.cache->server->module_config, &cache_module);
    
    /* write headers */
    file_cache_write_hdrs(cq);
    
    /* move the data over */
    if(cq->state & FC_TEMP_FILE && cq->fp)
    {
        if(!cq->datafile) cq->datafile = data_file(cq->el.cache, cq->el.name);
        if(unlink(cq->datafile)) /* if we can remove it, we clearly don't have to build the dirs */
            mkdir_structure(cq->datafile, conf->root);

#if defined(OS2) || defined(WIN32)         /* Under OS/2 use rename. */
        if (rename(cq->tempfile, cq->datafile) == -1)
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, el->cache->server,
                         "proxy: error renaming cache file %s to %s",
                         c->tempfile, cq->datafile);
#else
        if (link(cq->tempfile, cq->datafile) == -1)
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, el->cache->server,
                         "proxy: error linking cache file %s to %s",
                         cq->tempfile, cq->datafile);
        if (unlink(cq->tempfile) == -1)
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, el->cache->server,
                         "proxy: error deleting temp file %s", cq->tempfile);
#endif
    }
    if(cq->fp)
        apr_close(cq->fp); /* if you finalize, you are done writing, so close it */

    memset(el, '\0', sizeof(ap_cache_el)); /* clear it out, for the sake of clarity */
    return APR_SUCCESS;
}
